# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Defines the logger for the search submodule.
"""

import logging

SEARCH_LOGGER = logging.getLogger('nodefinder.search')
