# -*- coding: utf-8 -*-

# © 2017-2019, ETH Zurich, Institut für Theoretische Physik
# Author: Dominik Gresch <greschd@gmx.ch>
"""
Submodule for identifying nodal features such as points and lines from a
point-cloud of nodes as created by the :mod:`.search` submodule.
"""

from ._run import *
from . import result
from . import plot

__all__ = ['result', 'plot'] + _run.__all__  # pylint: disable=undefined-variable
